package rootgui.controller;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

public class RunnableBlockingQueueList {

	/*
	 * List of Threads that are consuming the Runnable tasks placed in the
	 * executionQueues (one for each Simulation client)
	 */
	private Map<Integer, Thread> executionThreads;

	/*
	 * List of BlockingQueue<Runnable> objects, one for each execution thread
	 */
	private Map<Integer, BlockingQueue<Runnable>> executionQueues;

	// private List<Boolean> threadAlive;

	/*
	 * Counters for each queue, indicating how many runnables they have executed
	 * (since the counters were last reset)
	 */
	private Map<Integer, Integer> executionCount;

	private List<RunnableBlockingQueueListActionListener> listeners;

	/*
	 * Name of RunnableBlockingQueueList
	 */
	private String name;

	/*
	 * 
	 */
	public RunnableBlockingQueueList(String name) {
		executionQueues = new HashMap<Integer, BlockingQueue<Runnable>>();
		executionThreads = new HashMap<Integer, Thread>();
		// threadAlive = new ArrayList<Boolean>();
		executionCount = new HashMap<Integer, Integer>();
		listeners = new ArrayList<RunnableBlockingQueueListActionListener>();
		this.name = name;
	}

	public RunnableBlockingQueueList(String name,
			RunnableBlockingQueueListActionListener listener) {
		this(name);
		listeners.add(listener);
	}

	public void addListener(RunnableBlockingQueueListActionListener l) {
		listeners.add(l);
	}

	public void removeListener(RunnableBlockingQueueListActionListener l) {
		listeners.remove(l);
	}

	public String getName() {
		return name;
	}

	/*
	 * Appends another queue to the Queue list. It returns the index of the new
	 * queue
	 */
	public synchronized int createQueue(final int simID) {
		// creating queue that will receive Runnable tasks

		final BlockingQueue<Runnable> bq = new LinkedBlockingQueue<Runnable>();
		executionQueues.put(simID, bq);
		executionCount.put(simID, new Integer(0));

		// creating consumer thread
		Thread t = new Thread(new Runnable() {
			@Override
			public String toString() {
				return name + ": Execution thread for queue <" + simID + ">";
			}

			@Override
			public void run() {
				try {
					while (bq != null) {
						// BlockingQueue<Runnable> bq =
						Runnable r = bq.take();// bq.poll(1, TimeUnit.SECONDS);
						// if (r != null) {
						// System.out.println("controller queue [" + srIndex +
						// "]: size = " + serverCommQueues.get(srIndex).size());
						// System.err.println(name + ": Starting runnable: " +
						// r);
						int jobID = executionCount.get(simID);
						Thread t = new Thread(r);
						t.start();
						t.join();
						try {
							executionCount.put(simID, jobID + 1);
						} catch (Exception e) {
							System.err.println(e.getMessage());
						}
						// System.err.println(name + ": Finished runnable [" +
						// jobID + "]: " + r);
						for (RunnableBlockingQueueListActionListener l : listeners)
							l.fireJobFinished(simID, jobID, bq.size());
					}
					// }
				} catch (InterruptedException ex) {
					// just terminates the thread by exiting its run method					
					return;
				} catch (Exception ex){
					System.err.println("Exception in RBQL, execution thread: " + ex.getMessage());
					return;
				}
			}
		});
		executionThreads.put(simID, t);
		t.start();
		return simID;
	}

	/*
	 * Terminates and removes an execution queue
	 */
	public synchronized void removeQueue(final int simID) {
		executionThreads.get(simID).interrupt();
		executionThreads.remove(simID);

		executionQueues.get(simID).clear();
		executionQueues.remove(simID);

		executionCount.remove(simID);
	}

	public synchronized void removeAllQueues() {
		for (int simID : executionQueues.keySet())
			removeQueue(simID);
	}

	public int addRunnableToQueue(final int simID, Runnable r) {
		try {
			int id = executionQueues.get(simID).size();
			executionQueues.get(simID).put(r);
			for (RunnableBlockingQueueListActionListener l : listeners)
				l.fireJobAdded(simID, id, executionQueues.get(simID).size());
			// System.out.println("controller queue [" + srIndex +
			// "]: size = " + serverCommQueues.get(srIndex).size());

			return id;
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		return -5;
	}

	/*
	 * Puts a Runnable object in the execution queue identified by srIndex.
	 * Returns the position (index) in the queue, where the Runnable object was
	 * appended.
	 */
	public int addBlockingRunnableToQueue(final int simID, Runnable r) {
		if (executionQueues.containsKey(simID))
			try {
				final int id = executionQueues.get(simID).size();
				final int initialExecutionCount = executionCount.get(simID);
				executionQueues.get(simID).put(r);
				for (RunnableBlockingQueueListActionListener l : listeners)
					l.fireJobAdded(simID, id, id + 1);
				// System.out.println("controller queue [" + srIndex +
				// "]: size = " + serverCommQueues.get(srIndex).size());

				Thread t = new Thread(new Runnable() {

					@Override
					public void run() {
						int currentExecutionCount = executionCount.get(simID);
						while (currentExecutionCount <= initialExecutionCount + id) {
							try {
								Thread.sleep(100);
								currentExecutionCount = executionCount.get(simID);
							} catch (InterruptedException e) {
								// TODO Auto-generated catch block
//								e.printStackTrace();								
							} catch (Exception e){
								return;
							}
						}
					}
				});
				t.start();
				try {
					t.join();
				} catch (InterruptedException e) {
					t.interrupt();
				}

				return id;
			} catch (InterruptedException e) {
				
			}
		return -5;
	}

	/*
	 * Does not return until the designated serverCommQueue is empty
	 */
	// public void blockTillQueueEmpty(final int srIndex) {
	// Thread t = new Thread(new Runnable() {
	//
	// @Override
	// public void run() {
	// while (serverCommQueues.get(srIndex).size() != 0) {
	// try {
	// Thread.sleep(100);
	// } catch (InterruptedException e) {
	// // TODO Auto-generated catch block
	// e.printStackTrace();
	// }
	// }
	// }
	// });
	// t.start();
	// try {
	// t.join();
	// } catch (InterruptedException e) {
	// // TODO Auto-generated catch block
	// e.printStackTrace();
	// }
	// }

	public synchronized int getQueuedJobsCount(int srIndex) {
		try {
			return executionQueues.get(srIndex).size();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			// e.printStackTrace();
			return 0;
		}
	}

	public synchronized int getQueuedJobsCount() {
		int sum = 0;
		for (int i = 0; i < executionQueues.size(); i++)
			sum += executionQueues.get(i).size();
		return sum;
	}

//	public static void main(String[] args) {
//		RunnableBlockingQueueList rbql = new RunnableBlockingQueueList(
//				"testRBQL");
//		int srIndex = rbql.createQueue(0);
//
//		Runnable r = new Runnable() {
//
//			@Override
//			public void run() {
//				try {
//					Thread.sleep(2000);
//				} catch (InterruptedException e) {
//					// TODO Auto-generated catch block
//					e.printStackTrace();
//				}
//				System.out.println(this + " finished...");
//			}
//		};
//
//		rbql.addRunnableToQueue(srIndex, r);
//
//		rbql.addRunnableToQueue(srIndex, r);
//
//		rbql.addBlockingRunnableToQueue(srIndex, r);
//
//		rbql.removeAllQueues();
//	}
}
