package rootgui.model;

import java.awt.Rectangle;
import java.io.Serializable;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import rootgui.model.objects.*;
import rootgui.model.parser.CellsetGeometryParser;

/*
 * Class representing the geometry of a tissue
 */
public class Geometry implements Cloneable, Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 625744756782972834L;
	/*
	 * Hashtable of Cell objects, the key being the cell id and the value being the cell.
	 */
	private Hashtable<Integer, Cell> cells;
	/*
	 * Hashtable of Wall objects, the key being the wall id and the value being the wall.
	 */
	private Hashtable<Integer, Wall> walls;
	/*
	 * 
	 */
	private Hashtable<Integer, String> cellGroups, wallGroups;

	/*
	 * True if the geometry has changed;
	 */
	private boolean dirty = false;

	private Rectangle boundingRectangle;

	public Geometry() {
		cells = new Hashtable<Integer, Cell>();
		walls = new Hashtable<Integer, Wall>();
		cellGroups = new Hashtable<Integer, String>();
		wallGroups = new Hashtable<Integer, String>();
	}

	public Geometry(Geometry render) {
		cells = new Hashtable<Integer, Cell>();
		walls = new Hashtable<Integer, Wall>();
		for (Cell cell : render.getCells()) {
			addCell(cell);
		}
		cellGroups = (Hashtable<Integer, String>) render.cellGroups.clone();
		wallGroups = (Hashtable<Integer, String>) render.wallGroups.clone();
		dirty = true;
	}

	public void addCell(Cell cell) {
		cells.put(cell.getId(), cell);
		for (Wall w : cell.getWalls()) {
			walls.put(w.getId(), w);
		}
		dirty = true;
	}

	public Collection<Cell> getCells() {
		return cells.values();
	}

	public int getCellCount() {
		return cells.size();
	}

	public Collection<Wall> getWalls() {
		return walls.values();
	}

	public Cell getCellById(int id) {
		return cells.get(id);
	}

	public void addCellGroupName(Integer cellGroupId, String cellGroupName) {
		cellGroups.put(cellGroupId, cellGroupName);
	}

	public String getCellGroupNameById(Integer groupId) {
		return cellGroups.get(groupId);
	}

	public void addWallGroupName(Integer wallGroupId, String wallGroupName) {
		wallGroups.put(wallGroupId, wallGroupName);
	}

	public String getWallGroupNameById(Integer wallGroupId) {
		return wallGroups.get(wallGroupId);
	}

	public static Geometry parseXMLString(String xmlString) {
		CellsetGeometryParser cgp = new CellsetGeometryParser();
		return cgp.getGeometryFromString(xmlString);
	}

	public int getMinY() {
		return (int) getBoundingRectangle().getMinY();
	}

	public int getMinX() {
		return (int) getBoundingRectangle().getMinX();
	}

	public int getTissueWidth() {
		return (int) getBoundingRectangle().getWidth();
	}

	public int getTissueHeight() {
		return (int) getBoundingRectangle().getHeight();
	}

	public void setGeometryChanged(boolean changed){
		dirty = changed;
	}
	
	public Rectangle getBoundingRectangle() {
		if (dirty) {
			boundingRectangle = new Rectangle();

			int xmin = Integer.MAX_VALUE;
			int xmax = 0;
			int ymin = Integer.MAX_VALUE;
			int ymax = 0;
			for (Cell cell : cells.values()) {
				int[] xPoints = cell.getXpoints();
				int[] yPoints = cell.getYpoints();
				for (int i = 0; i < xPoints.length; i++) {
					if (xPoints[i] < xmin)
						xmin = xPoints[i];
					if (xPoints[i] > xmax)
						xmax = xPoints[i];
					if (yPoints[i] < ymin)
						ymin = yPoints[i];
					if (yPoints[i] > ymax)
						ymax = yPoints[i];
				}
			}
			boundingRectangle.setBounds(xmin, ymin, xmax - xmin, ymax - ymin);
			dirty = false;
		}
		return (Rectangle)boundingRectangle.clone();
	}

	@Override
	public Object clone() throws CloneNotSupportedException {
		return super.clone();
	}

}
