package rootgui.model.recordstore;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import rootgui.model.Geometry;
import rootgui.model.objects.Species;
import rootgui.model.parser.CellsetGeometryParser;
import rootgui.model.parser.D2CSVparser;

public class Record implements Serializable {
	/**
	 * 
	 */
	private static final long serialVersionUID = -2057481603692660047L;

	public static final int NO_ITERATION = -1;
	/*
	 * 
	 */
	private Geometry geometry;

	/*
	 * Map containing species records associated with cell ids. Keys are the
	 * species type, whereas the values are maps of species and species records.
	 */
	private Map<Species.Type, Map<Species, SpeciesRecord>> data;

	protected int iteration;

	protected IRecordStore store;

	/*
	 * Map containing species records associated with edge ids. Keys are the
	 * species name, whereas the values are the species records.
	 */
	// private Map<String, SpeciesRecord> edgeData;

	public Record(IRecordStore store) {
		geometry = new Geometry();
		data = new HashMap<Species.Type, Map<Species, SpeciesRecord>>();
		iteration = NO_ITERATION;
		this.store = store;
//		if (store == null)
//			System.err.println("Record(" + store + ") incorrect as store == null.");
	}

//	public Record(Geometry g) {
//		geometry = g;
//		data = new HashMap<Species.Type, Map<String, SpeciesRecord>>();
//		iteration = NO_ITERATION;
//		store = null;
//	}

	public void setIteration(int iteration) {
		this.iteration = iteration;
	}

	public int getIteration() {
		return iteration;
	}

	public void setStore(IRecordStore store, int iteration) {
		this.store = store;
		if (store == null)
			System.err.println("Record.setStore(" + store + ", " + iteration +") incorrect as store == null.");
		this.iteration = iteration;
	}

	public void detachFromStore() {
		this.store = null;
		this.iteration = NO_ITERATION;
	}

	public IRecordStore getStore() {
		return store;
	}

	public Geometry getGeometry() {
		return geometry;
	}

	public void setGeometry(Geometry geometry) {
		this.geometry = geometry;
		fireUpdated();
	}

	public void setGeometryFromFile(String cellSetFile) {
		CellsetGeometryParser cgp = new CellsetGeometryParser();
		geometry = cgp.getGeometryFromFile(cellSetFile);
		fireUpdated();
	}

	public void setRecord(SpeciesRecord sr) {
		Map<Species, SpeciesRecord> m = null;
		Species species = sr.getSpecies();
		if (!data.containsKey(species.getType())) {
			m = new HashMap<Species, SpeciesRecord>();
			data.put(species.getType(), m);
		} else {
			m = data.get(sr.getSpecies().getType());
		}
		m.put(species, sr);
		fireUpdated();
	}

	/*
	 * @param speciesName: the name of the species
	 */
	public void setRecord(Species species, Map<String, Double> values) {
		Map<Species, SpeciesRecord> m = null;
		if (!data.containsKey(species.getType())) {
			m = new HashMap<Species, SpeciesRecord>();
			data.put(species.getType(), m);
		} else {
			m = data.get(species.getType());
		}
		m.put(species, new SpeciesRecord(species, values, this));
		fireUpdated();
	}

	public Map<String, Double> getSpeciesData(Species species) {
		return data.get(species.getType()).get(species).getValues();
	}

	/*
	 * Returns a map of all the SpeciesRecords objects within the Record
	 */
	public Map<Species, SpeciesRecord> getSpeciesRecordByType(Species.Type type) {
		return data.get(type);
	}

	public void setCellRecordFromCSV(Species species, String csvFile) {
		setCellRecordFromCSV(species, csvFile, 0);
		fireUpdated();
	}

	public void setCellRecordFromCSV(Species species, String csvFile,
			int iteration) {
		D2CSVparser dp = new D2CSVparser(csvFile);
		SpeciesRecord sr = new SpeciesRecord(species, dp.getDataRow(iteration),
				this);
		Map<Species, SpeciesRecord> m = data.get(species.getType());
		if (m == null) {
			m = new HashMap<Species, SpeciesRecord>();
			data.put(species.getType(), m);
		}
		m.put(species, sr);
		fireUpdated();
	}

	/*
	 * Returns a Set<SpeciesRecord> containing all the SpeciesRecord objects of
	 * the Record
	 */
	public Set<SpeciesRecord> getData() {
		Set<SpeciesRecord> result = new HashSet<SpeciesRecord>();
		for (Species.Type st : data.keySet()) {
			for (SpeciesRecord sr : data.get(st).values()) {
				result.add(sr);
			}
		}
		return result;
	}

	public SpeciesRecord getData(Species species) {
		try {
			return data.get(species.getType()).get(species);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			// e.printStackTrace();
			return null;
		}
	}

//	/*
//	 * Returns the SpeciesRecord object corresponding to a species name
//	 */
//	public SpeciesRecord getCellData(String speciesName) {
//		if (data != null) {
//			if (data.containsKey(Species.Type.CELL)
//					&& data.get(Species.Type.CELL).containsKey(speciesName)) {
//				return data.get(Species.Type.CELL).get(speciesName);
//			} else if (data.containsKey(Species.Type.CELL_WALL)
//					&& data.get(Species.Type.CELL_WALL)
//							.containsKey(speciesName)) {
//				return data.get(Species.Type.CELL_WALL).get(speciesName);
//			}
//		}
//		return null;
//	}

	/*
	 * Returns a Set of species names for which data is stored in this Record
	 */
	public Set<String> getSpeciesNames() {
		return getSpeciesNames(null);
	}

	/*
	 * Returns a Set of species names, of a specified species type, for which
	 * data is stored in this Record
	 */
	public Set<String> getSpeciesNames(Species.Type speciesType) {
		if (data != null) {
			if (speciesType != null){
				Set<String> al = new TreeSet<String>();
				for (Species s : data.get(speciesType).keySet())
					al.add(s.getName());
				return al;
			}
			else {
				Set<String> al = new TreeSet<String>();
				for (Species.Type st : data.keySet()) {
					for (Species s : data.get(st).keySet())
						al.add(s.getName());
				}
				return al;
			}
		}
		return new TreeSet<String>();
	}

	/*
	 * Returns a Set of species for which data is stored in this Record
	 */
	public Set<Species> getSpecies() {
		if (data != null) {
			Set<Species> al = new TreeSet<Species>();
			for (Species.Type st : data.keySet()) {
				for (Species s : data.get(st).keySet())
					al.add(s);
			}
			return al;
		}
		return new TreeSet<Species>();
	}

	// public SpeciesRecord getWallData(String speciesName) {
	// return data.get(Species.Type.WALL).get(speciesName);
	// }

//	public SpeciesRecord getCellWallData(String speciesName) {
//		return data.get(Species.Type.CELL_WALL).get(speciesName);
//	}

	public void fireCellDataUpdated(String cellID, String speciesName, double value){
		if (store != null) {
			store.fireCellDataUpdated(this, cellID, speciesName, value);
		} 
//		else 
//			System.err.println("Record: fireCellDataUpdated not applied as store == null.");
	}
	
	void fireUpdated() {
		if (store != null) {
			store.fireRecordUpdated(iteration);
		} 
//		else 
//			System.err.println("Record: fireUpdated not applied as store == null.");
	}

	@Override
	public String toString() {
		return "Record " + iteration + ": geometry = "
				+ (geometry != null ? geometry.toString() : "null")
				+ ", cellData = " + (data != null ? data.toString() : "null");
	}
}
