package rootgui.model.recordstore;

import java.io.Serializable;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Arrays;

import rootgui.model.objects.Species;

/*
 * A SpeciesRecord stores all data about a species 
 * (in relation to cell or edge ids), for a given iteration
 */
public class SpeciesRecord implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 8509056915185374729L;

	/*
	 * A reference to the Record object in which this SpeciesRecord is to be
	 * stored
	 */
	private Record record;

	/*
	 * A map of data, where keys are ids and values are the species values
	 */
	private Map<String, Double> data;

	/*
	 * The species name
	 */
	private Species species;

	// public SpeciesRecord(Species species, double[] cellSpeciesValues,
	// Record record) {
	// this(species, cellSpeciesValues, record);
	// }

	/*
	 * Constructor useful for cell species data
	 */
	public SpeciesRecord(Species species, double[] speciesValues, Record record) {
		this.species = species;
		data = new HashMap<String, Double>();
		for (int i = 0; i < speciesValues.length; i++)
			data.put("" + i, speciesValues[i]);
		this.record = record;
	}

	public SpeciesRecord(Species species, Map<String, Double> speciesValues,
			Record record) {
		this.species = species;
		data = speciesValues;
		this.record = record;
	}

	public void put(Integer elementId, Double value) {
		String key = elementId.toString();
		if (!data.containsKey(key) || !data.get(key).equals(value)) {
			data.put(key, value);
			record.fireCellDataUpdated(key, species.getName(), value);
		}
	}

	public Double get(String id) {
		return data.get(id);
	}

	public Double get(Integer elementId) {
		return data.get(elementId.toString());
	}

	// public String getSpeciesName() {
	// return species.getName();
	// }

	public Species getSpecies() {
		return species;
	}

	public Map<String, Double> getValues() {
		return data;
	}

	public Double[] getValuesArray() {
		return data.values().toArray(new Double[] {});
	}

	public int size() {
		return data.size();
	}

	public Record getRecord() {
		return record;
	}

	public double getMaxValue() {
		double maxConcentration = 0;
		for (Double d : data.values()) {
			double concentration = d.doubleValue();
			if (maxConcentration < concentration)
				maxConcentration = concentration;
		}
		return maxConcentration;
	}
	
	/*
	 * Returns the maximum value within the SpeciesRecord
	 */
	public double getMaxValue(double percentile) {
		// this function returns a percentile of the concentration lists, given as a decimal between 0 and 1.
		double maxConcentration = 0;
		if (percentile < 0.1 || percentile > 1)
			return getMaxValue();
		int numCells = data.size();
		int index = (int) Math.ceil(percentile * numCells - 1);	
		Double[] valuesArray = getValuesArray();
		Arrays.sort(valuesArray);
		maxConcentration = valuesArray[index];
		if (maxConcentration == 0) 
			return getMaxValue();
		return maxConcentration;
	}

	/*
	 * Returns the minimum value within the SpeciesRecord
	 */
	public double getMinValue() {
		double minConcentration = 0;
		for (Double d : data.values()) {
			double concentration = d.doubleValue();
			if (minConcentration > concentration)
				minConcentration = concentration;
		}
		return minConcentration;
	}
}
