package rootgui.view.settingspanel;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.MouseEvent;
import java.text.NumberFormat;

import javax.swing.JComponent;
import javax.swing.JFormattedTextField;
import javax.swing.ToolTipManager;

import rootgui.model.Parameter;

/**
 * the parameter field that have functions including checking input text format
 * (i.e. checking if it is a number), displaying the unit of the field as a tool
 * tip, changing background color if the new value is different from default
 * 
 * @author Forest
 * 
 */
public class ParameterField extends JFormattedTextField {

	/**
	 * 
	 */
	private static final long serialVersionUID = -4633990787904318943L;


	/*
	 * name of this parameter field
	 */
	final private String name;


	/*
	 * default value of this parameter field
	 */
	private double defaultValue;


	double maxVal;


	double minVal;

	/*
	 * number of columns that the parameter value is allowed to display
	 */
	static final int LIMIT_OF_COLUMNS = 6;

	/*
	 * format for the parameter value
	 */
	private static NumberFormat format;

	/*
	 * background color used if the parameter field is filled with the default
	 * value
	 */
	private static Color defaultColor;

	/*
	 * background color used if the parameter field is filled with a value
	 * different from defaut
	 */
	private static Color modificationColor;

	public ParameterField(String name) {
		this(name, null, -1e30, 1e30);
	}

	public Parameter getParameter() {
		double val = Double.parseDouble(getText());
		if(val<minVal)
			setText(format.format(minVal));
		if(val>maxVal)
			setText(format.format(maxVal));
		return new Parameter(name, getText());
	}

	public ParameterField(String name, final String units, final double minVal, final double maxVal) {
		super(format = NumberFormat.getNumberInstance());
		this.name = name;
		this.maxVal = maxVal;
		this.minVal = minVal;
		
		format.setMaximumFractionDigits(3);
		format.setMinimumFractionDigits(1);
		setColumns(LIMIT_OF_COLUMNS - 2);
		defaultColor = Color.WHITE;
		modificationColor = Color.LIGHT_GRAY;
		addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				try {
					ParameterField f = (ParameterField) e.getSource();
					double val = Double.parseDouble(f.getText());
					if(val<minVal)
						f.setText(format.format(minVal));
					if(val>maxVal)
						f.setText(format.format(maxVal));
					
					f.updateBackground();

				} catch (NumberFormatException nfe) {
					nfe.printStackTrace();
				}	
			}
		});
		
		addFocusListener(new FocusListener() {

			@Override
			public void focusLost(FocusEvent e) {
				if (e.isTemporary()) {
					return;
				}
				
				try {
					ParameterField f = (ParameterField) e.getSource();
					double val = Double.parseDouble(f.getText());
					if(val<minVal)
						f.setText(format.format(minVal));
					if(val>maxVal)
						f.setText(format.format(maxVal));
					
					f.updateBackground();

				} catch (NumberFormatException nfe) {
					nfe.printStackTrace();
				}

				// hide the tool tip when losing focus
				JComponent component = (JComponent) e.getSource();
				MouseEvent phantom = new MouseEvent(component,
						MouseEvent.MOUSE_MOVED, System.currentTimeMillis(), 0,
						0, 0, 0, false);
				ToolTipManager.sharedInstance().mouseExited(phantom);
			}

			@Override
			public void focusGained(FocusEvent e) {
				JComponent component = (JComponent) e.getSource();
				component.setToolTipText(units);

				MouseEvent phantom = new MouseEvent(component,
						MouseEvent.MOUSE_MOVED, System.currentTimeMillis(), 0,
						component.getWidth(), 0,// component.getHeight()/3,
						0, false);

				ToolTipManager.sharedInstance().mouseMoved(phantom);
			}
		});

	}

	public void updateBackground(){
		try {
			if (Double.parseDouble(getText().replace(",", "")) != defaultValue) {
				setBackground(modificationColor);
			} else {
				setBackground(defaultColor);
			}
		} catch (NumberFormatException e) {
			// TODO Auto-generated catch block
			//e.printStackTrace();
			// if text =="" then deal gracefully with parse exception
			setBackground(defaultColor);
		}
		
	}

	public static double roundToSignificantFigures(double num, int n) {
		if (num == 0) {
			return 0;
		}

		final double d = Math.ceil(Math.log10(num < 0 ? -num : num));
		final int power = n - (int) d;

		final double magnitude = Math.pow(10, power);
		final long shifted = Math.round(num * magnitude);
		return shifted / magnitude;
	}

	public void setToDefaultValue() {
		setText(format.format(defaultValue));
		setBackground(defaultColor);
	}

	public void setDefaultValue(double v) {
//		System.out.println("Default "+name+" "+defaultValue+" "+v);
		defaultValue = v; //Double.parseDouble(format.format(v).replaceAll(",", ""));
	}

	public double getDefaultValue() {
		return defaultValue;
	}

	public String getName() {
		return name;
	}
}
