package visualiser.viewpanel.view;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.io.File;

import javax.imageio.ImageIO;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JSpinner;
import javax.swing.Timer;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import rootgui.model.objects.Species;
import rootgui.model.recordstore.Record;

import visualiser.Figure;
import visualiser.IVisualiser;
import visualiser.viewpanel.controller.CDPController;
import visualiser.viewpanel.view.tabs.VisualiserTab;

/**
 * 
 * @author Lei Yu
 * 
 */
public class VisualiserPanel extends JPanel implements ActionListener {
	public final static double ZOOM_INCREMENT = 0.05;
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private CellDrawingPanel mCdp;
	private ControlBar mBar;

	private CDPController mCdpController;

	private Timer mTimer;
	private final int mPlayDelay = 500;

	/*
	 * Specifies what was the biggest rendered record
	 */
	private int mCurrentMaxIteration = -1;
	private int mIteration = 0;
	private int mMaxIteration = 0;

	/*
	 * Index within a VisualiserTab
	 */
	private int mIndex = -1;

	/*
	 * Index of species being rendered (from within the Figure object)
	 */
	// private int figureIndex;

	/*
	 * Figure used for rendering information in the VisualiserTab
	 */
	// private Figure figure;

	/*
	 * Reference to IVisualiser object that contains VisualiserTabs
	 */
	private IVisualiser mVis;

	/*
	 * Reference to VisualiserTab object that contains the VisualiserPanel
	 */
	private VisualiserTab mVisTab;

	/*
	 * Specifies if the VP should synchronise or not with the other VPs (e.g.
	 * when the rendered iteration is changed or when the Play button is
	 * pressed)
	 */
	private boolean mViewSynchronised = true;

	/*
	 * @param id: index number within the VisualiserTab
	 * 
	 * @param f: figure
	 */
	public VisualiserPanel(int id, Figure f, int figureIndex, IVisualiser vis,
			CellDrawingPanel cdp, CDPController controller, VisualiserTab visTab) {

		this.mIndex = id;
		// figure = f;
		this.mVis = vis;
		this.mCdp = cdp;
		this.mCdpController = controller;
		mBar = new ControlBar();
		this.mVisTab = visTab;
	}

	/*
	 * 
	 */
	public VisualiserPanel(int id, IVisualiser vis, CellDrawingPanel cdp,
			CDPController controller, VisualiserTab visTab) {

		this.mIndex = id;
		// figure = f;
		this.mVis = vis;
		this.mCdp = cdp;
		this.mCdpController = controller;
		this.mVisTab = visTab;
		initComponent();
	}

	private void initComponent() {
		// setClosable(true);
		// JScrollPane spane = new JScrollPane(cdp);
		mCdp.setPreferredSize(new Dimension(750, 400));
		setLayout(new BorderLayout());
		add(mCdp, BorderLayout.CENTER);

		// Set up a timer that calls this object's action handler.
		mTimer = new Timer(mPlayDelay, this);
		mTimer.setInitialDelay(100);
		// timer.setCoalesce(true);

		mBar = new ControlBar();
		add(mBar, BorderLayout.SOUTH);
		mBar.addExportActionListener(new ExportActionListener());
		mBar.addZoomActionListener(new ZoomActionListener());
		mBar.addIterationSliderChangeListner(new IterationSliderChangeListener());
		mBar.addSpeciesSliderChangeListener(new SpeciesSliderChangeListener());
		mBar.addColourBarSelectionActionListener(new ColourBarSelectionListener());
		mBar.addPlayActionListener(new PlayActionListener());
		mBar.addZoomFactorChangeListener(new SpinnerChangeListener());

		setVisible(true);
	}

	public void setViewSynchronised(boolean onOff) {
		mViewSynchronised = onOff;
	}

	public boolean isViewSynchronised() {
		return mViewSynchronised;
	}

	public void setFigure(Figure f, int index) {
		mCdp.setFigure(f, index);
		mBar.setSpecies(f.getSpecies(), index);
	}

	public void setLoadedSpecies(Species[] species) {
		mBar.setSpecies(species, mBar.getSelectedSpeciesIndex());
	}

	public void setZoomRatio(double zoomRatio) {
		mCdp.setZoomRatio(zoomRatio);
	}

	public void setFitZoomRatio() {
		mCdp.setFitZoomRatio();
	}

	public Species getSelectedSpecies() {
		return mBar.getSelectedSpecies();
	}

	public void setZoomValue(double ratio) {
		mBar.setZoomFactor(ratio);
	}

	public double getZoomRatio() {
		return mCdp.getZoomRatio();
	}

	/*
	 * Returns the index of the VisualiserPanel, within the VisualiserTab
	 */
	public int getIndex() {
		return mIndex;
	}

	public void exportPNG(File file) {
		BufferedImage bi = new BufferedImage(mCdp.getWidth(), mCdp.getHeight(),
				BufferedImage.TYPE_INT_ARGB);
		Graphics g = bi.createGraphics();
		mCdp.paint(g);
		g.dispose();
		try {
			ImageIO.write(bi, "png", file);
		} catch (Exception e) {

		}
	}

	public void reset() {
		if (mCdpController != null)
			mCdpController.setRecord(null, true);
		mCurrentMaxIteration = -1;
		mMaxIteration = 0;
		mBar.setButtonsEnabled(false);
		mBar.setPlayButtonEnabled(false);
		mBar.setIteration(0);
		mBar.setMaxIterations(0);
	}
	
	/*
	 * @param Record r: the Record object with the data that needs visualising
	 * 
	 * @param boolean setControlBar: specifies if the ControlBar should be
	 * modified by this action
	 */
	public void setRecord(Record r, boolean setControlBar, boolean initial) {
		// record = r;
		if (mCdpController != null)
			mCdpController.setRecord(r, initial);

		if (r != null) {
			if (r.getIteration() > mCurrentMaxIteration)
				mCurrentMaxIteration = r.getIteration();
		} else {
			mCurrentMaxIteration = -1;
			mMaxIteration = 0;
		}

		if (setControlBar) {
			mBar.setButtonsEnabled(r != null);
			mBar.setPlayButtonEnabled((!initial && mCurrentMaxIteration > 0)
					|| ((mCdpController.getRecord() != null && mCdpController.getRecord().getStore()!=null) ? mCdpController
							.getRecord().getStore().size() > 1 : false));
			int iteration = r != null ? r.getIteration() : 0;

			if (mCurrentMaxIteration > mBar.getMaxIterations())
				mBar.setMaxIterations(mCurrentMaxIteration);
			else if (r == null)
				mBar.setMaxIterations(0);

			mBar.setIteration(iteration);
			if (iteration == 0 && r != null) {
				mBar.setZoomFactor(mCdp.getZoomRatio());
			}
		}

	}

	// public boolean isStepAll() {
	// return bar.isStepAll();
	// }

	public void setMaxIteration(int i) {
		if (i != mMaxIteration) {
			mMaxIteration = i;
			mBar.setMaxIterations(i);
		}
	}

	class PlayActionListener implements ActionListener {

		@Override
		public void actionPerformed(ActionEvent e) {
			if (e.getActionCommand().equals("Play") && !mTimer.isRunning()) {
				// visTab.setStartButton(true);
				// currentIteration = visTab.getCurrentIteration();
				// iteration = currentIteration;
				mIteration = 0;
				// System.err.println("The current Iteration is " +
				// currentIteration);
				mTimer.start();
				// System.err.println("timer started");
			}

		}

	}

	class ExportActionListener implements ActionListener {

		@Override
		public void actionPerformed(ActionEvent e) {
			// if (e.getActionCommand().equals("Export")) {
			JComboBox cb = (JComboBox) e.getSource();
			// visTab.setStartButton(true);
			JFileChooser fileChooser = new JFileChooser();
			int selectedIndex = cb.getSelectedIndex();
			if (selectedIndex == 0) {
				mCdp.setExporting(true);
				int returnVal = fileChooser.showSaveDialog(mCdp);
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					File file = fileChooser.getSelectedFile();
					String path = file.getAbsolutePath();
					File newFile = new File(path + ".png");
					exportPNG(newFile);
				}
			} else {
				mCdp.setExporting(true);
				fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				int returnVal = fileChooser.showSaveDialog(mCdp);

				if (returnVal == JFileChooser.APPROVE_OPTION) {
					File file = fileChooser.getSelectedFile();
					String path = file.getAbsolutePath();
					System.err.println(path);
					if (file.mkdir())
						;
					// System.err.println("directory created");
					// else
					// System.err.println("directory not created");

					int currentIteration = mBar.getCurrentIteration();
					if (currentIteration == 0)
						currentIteration = mMaxIteration;
					// int i = currentIteration;

					for (int i = currentIteration; i >= 0; i--) {
						// bar.getSlider().setValue(i);
						mVisTab.setRenderedIteration(i);
						File newfile = new File(path + "/"
								+ getSelectedSpecies().getPublicName()
								+ "_step_" + i + ".png");
						exportPNG(newfile);
					}

					mVisTab.setRenderedIteration(currentIteration);
					// bar.getSlider().setValue(currentIteration);
				}
			}
			mCdp.setExporting(false);

		}

	}

	class SpinnerChangeListener implements ChangeListener {

		@Override
		public void stateChanged(ChangeEvent e) {
			JSpinner spinner = (JSpinner) e.getSource();
			if (spinner != null) {
				double zoomFactor = getZoomFactor(spinner);
				if (zoomFactor != mCdp.getZoomRatio()) {
					mVisTab.applyTransform(-1, zoomFactor, 0, 0, true);
					mVisTab.setSpinnerValue();
				}
			}
		}

	}

	class ZoomActionListener implements ActionListener {

		@Override
		public void actionPerformed(ActionEvent e) {
			if (e.getActionCommand().equals("-")) {
				// VisualiserTab vt = getActiveVisualiserTab();

				double tx = (mVisTab.getWidth() / 2) * ZOOM_INCREMENT;
				double ty = (mVisTab.getHeight() / 2) * ZOOM_INCREMENT;
				mVisTab.applyTransform(-1, 1 - ZOOM_INCREMENT, tx, ty, false);
				mVisTab.setSpinnerValue();

			} else if (e.getActionCommand().equals("+")) {
				// VisualiserTab vt = getActiveVisualiserTab();
				double tx = mVisTab.getWidth() / 2 * -ZOOM_INCREMENT;
				double ty = mVisTab.getHeight() / 2 * -ZOOM_INCREMENT;
				mVisTab.applyTransform(-1, 1 + ZOOM_INCREMENT, tx, ty, false);
				mVisTab.setSpinnerValue();
			}
		}
	}

	class IterationSliderChangeListener implements MouseMotionListener {
		int oldIteration = -1;

		@Override
		public void mouseMoved(MouseEvent e) {
			// TODO Auto-generated method stub

		}

		@Override
		public void mouseDragged(MouseEvent e) {
			JSlider source = (JSlider) e.getSource();
			// if (!source.getValueIsAdjusting()) {
			int iteration = (int) source.getValue();
			if (oldIteration != iteration && iteration <= mCurrentMaxIteration
					&& isViewSynchronised()) {
				oldIteration = iteration;
				mVisTab.setRenderedIteration(iteration, mIndex);
			}
		}
	}

	class SpeciesSliderChangeListener implements MouseMotionListener {
		int oldIndex = -1;

		public void mouseMoved(MouseEvent e) {
			// TODO Auto-generated method stub

		}

		@Override
		public void mouseDragged(MouseEvent e) {
			JSlider source = (JSlider) e.getSource();
			// if (!source.getValueIsAdjusting()) {
			int index = (int) source.getValue();
			if (index != oldIndex) {
				oldIndex = index;
				mCdp.setFigureIndex(index);
			}
		}
		// }

	}

	class ColourBarSelectionListener implements ActionListener {

		@Override
		public void actionPerformed(ActionEvent e) {
			if (mCdp != null) {
				JComboBox cb = (JComboBox) e.getSource();
				mCdp.setColourBar(cb.getSelectedIndex());
				/*
				 * if(cdp.getColourBarType()==0){
				 * cdp.setBackground(Color.WHITE); }
				 * if(cdp.getColourBarType()==1){
				 * cdp.setBackground(Color.BLACK); }
				 */
				mCdp.repaint();
			}
		}

	}

	public void applyTransform(double tZoom, double ttx, double tty,
			boolean isChangedBySpinner) {
		if (mCdp != null) {
			mCdp.applyTransform(tZoom, ttx, tty, isChangedBySpinner);
		}
	}

	public CDPController getController() {
		return mCdpController;
	}

	public CellDrawingPanel getView() {
		return mCdp;
	}

	public double getZoomFactor(JSpinner spinner) {
		double zoomFactor = (Double) spinner.getValue();
		return zoomFactor;
	}

	@Override
	public void actionPerformed(ActionEvent e) {

		// bar.getSlider().setValue(iteration);
		mVisTab.setRenderedIteration(mIteration, mIndex);

		if (mIteration >= mMaxIteration)
			mTimer.stop();
		else if (mIteration == mCurrentMaxIteration)
			mTimer.stop();
		mIteration++;
	}
}
