
"""
Module to import tissue in OpenAlea format (converted
from svg file), and do further processing:
rescaling the tissue, adding zero
PINs to edges which have none, adding AUX1 dictionary,
and classifying cell wall types according to angle
made with the x-axis
Also add rules for how to treat basic properties
on cell division.
"""


from openalea.celltissue import TissueDB
from openalea.tissueshape import edge_length,face_surface_2D, centroid
from openalea.scheduler import Scheduler,Task
from geom import area
from db_geom import ordered_wids_pids

from math import atan2, pi
import numpy

def import_tissue(fname):
    """
    Load OpenAlea tissue .zip file (made by create_tissue.py
    from svg).
    Generate initial mesh triangulation; extend properties
    that are defined on a subset of the edges/cells by
    create_tissue.py (i.e. PIN, border) to all of the
    things of that class, and set up the divided_props
    structure that informs the division algorithm how
    to update properties at division.
    """

    db=TissueDB()
    db.read(fname)

    from db_utilities import get_mesh

    db.set_property('scale', 1.0)
    db.set_description('scale', 'distance unit in microns')


    # Convert the position database to the plantGL Vector2 type
    #scale = 10.0
    scale = 1.0
    pos=db.get_property('position')
    pos=dict((pid, numpy.array(v)/scale) for pid, v in pos.iteritems())
    db.set_property('position', pos)
    return process_tissue(db)




def process_tissue(db):
    pos=db.get_property('position')
    t = db.tissue()
    cfg = db.get_config('config')

    #Read in properties
    graph = t.relation(cfg.graph_id) # directed cell to cell graph
    mesh = t.relation(cfg.mesh_id) # vertex,wall,cell topomesh

    # Fix PINs from tissue zip file
    if 'PIN' in db.properties():
        PIN = db.get_property('PIN')
    else:
    	PIN = {}
    PIN = dict((eid, PIN.get(eid, 0.0)) for eid in graph.edges()) # dictionary of pin concentrations for each directed graph edge
    db.set_property('PIN', PIN)

    # Triangulate all of the cells in the mesh

    # Fix border values in tissue file
    border = db.get_property('border')
    border = dict((cid, border.get(cid, None)) for cid in mesh.wisps(2))
    db.set_property('border', border)

    cell_type = db.get_property('cell_type')
    AUX = dict()
    for eid in graph.edges():
        sid = graph.source(eid)
        if cell_type[sid]==3 and border[sid]==7:
            AUX[eid] = .1
        else:
            AUX[eid] = 0
    db.set_property('AUX', AUX)
    db.set_description('AUX','')

    wall_types={}
    for cid in mesh.wisps(2):
        for wid, pid in zip(*ordered_wids_pids(mesh, pos, cid)):
            v1 = pos[pid]
            s = set(mesh.borders(1, wid))
            s.remove(pid)
            v0 = pos[s.pop()]
            angle = atan2(v1[1]-v0[1], v1[0]-v0[0])
            if -pi/4 <= angle < pi/4:
                wall_types[(cid, wid)] = 0
            if pi/4 <= angle < 3*pi/4:
                wall_types[(cid, wid)] = 1
            if 3*pi/4 <= angle or angle<-3*pi/4:
                wall_types[(cid, wid)] = 2
            if  -3*pi/4 <= angle < -pi/4:
                wall_types[(cid, wid)] = 3

    db.set_property('wall_types', wall_types)
    db.set_description('wall_types', ' wall annotations')

    #properties (loaded from tissue file) which updated on cell division
    divided_props={'cell_type': ('cell', 'attribute', -1),
                   'PIN': ('edge', 'concentration', 1),
                   'AUX': ('edge', 'concentration', .1),
                   'border': ('cell', 'attribute', 0),
                   'wall_types': ('(cell, wall)', 'attribute', 0),
                   }

    if 'LAX' in db.properties():
        divided_props['LAX'] = ('edge', 'concentration', 0)

    if 'VENUS_data_OAcellid' in db.properties():
        divided_props['VENUS_data_OAcellid'] = ('cell', 'concentration', 0)

    db.set_property('divided_props', divided_props)
    db.set_description('divided_props', 'property division rules')
  
    return db

